"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import defaultdict

from buttonwood.MarketObjects.Events import OrderEventConstants
from buttonwood.MarketObjects.Events.OrderEvents import AcknowledgementReport
from buttonwood.MarketObjects.Events.OrderEvents import CancelCommand
from buttonwood.MarketObjects.Events.OrderEvents import CancelReplaceCommand
from buttonwood.MarketObjects.Events.OrderEvents import CancelReport
from buttonwood.MarketObjects.Events.OrderEvents import ExecutionReport
from buttonwood.MarketObjects.Events.OrderEvents import FullFillReport
from buttonwood.MarketObjects.Events.OrderEvents import NewOrderCommand
from buttonwood.MarketObjects.Events.OrderEvents import PartialFillReport
from buttonwood.MarketObjects.Events.OrderEvents import RejectReport
from buttonwood.utils.IDGenerators import IDGenerator

import json


class CancelReplaceInfo(object):
    # TODO document class
    # TODO unit test

    def __init__(self, previous_exposure, new_exposure, side, market):
        self._prev = previous_exposure
        self._new = new_exposure
        mpi = market.product().mpi()
        if side.is_bid():
            self._price_delta = (self._new.price() - self._prev.price()) / mpi
        else:
            self._price_delta = (self._prev.price() - self._new.price()) / mpi
        self._qty_delta = new_exposure.qty() - previous_exposure.qty()

    def previous_exposure(self):
        return self._prev

    def new_exposure(self):
        return self._new

    def is_price_change(self):
        return self._price_delta != 0

    def is_better_price(self):
        return self._price_delta > 0

    def is_worse_price(self):
        return self._price_delta < 0

    def is_qty_change(self):
        return self._qty_delta != 0

    def is_qty_increase(self):
        return self._qty_delta > 0

    def is_qty_decrease(self):
        return self._qty_delta < 0

    def price_delta(self):
        return self._price_delta

    def qty_delta(self):
        return self._qty_delta

class Exposure(object):
    # TODO document this class

    def __init__(self, price, qty, causing_event_id):
        self._price = price
        self._qty = qty
        self._causing_event_id = causing_event_id

    def price(self):
        return self._price

    def qty(self):
        return self._qty

    def dec_qty(self, dec_amount):
        assert dec_amount > 0, "Must decrement qty with a positive integer."
        self._qty -= dec_amount

    def causing_event_id(self):
        return self._causing_event_id

    def equivalent_exposure(self, other):
        """
        This is to test that one exposure is equivalent to another. equivalent
         exposure is when the exposures are for the same price and the same qty.

        Not using __eq__ here because it does not matter if the causing_event_id
         is the same. So, equal exposures are equivalent but not all equivalent
         exposures are equal.

        If other is not an Exposure object then returns False.

        :param other: MarketObjects.Events.EventChains.Exposure
        :return: bool
        """
        if isinstance(other, Exposure):
            return self.price() == other.price() and self.qty() == other.qty()
        else:
            return False

    def __str__(self):
        return "%s: %d @ %s" % (str(self.causing_event_id()), self.qty(), str(self.price()))

    def __eq__(self, other):
        return self.price() == other.price() and \
               self.qty() == other.qty() and \
               self.causing_event_id() == other.causing_event_id()


class SubChain(object):
    # OPEN & CLOSE REASONS
    NEW_ORDER = 10  # open
    CANCEL_REPLACE_INCREASE_QTY = 20  # open and close
    CANCEL_REPLACE_PRICE = 30  # open and close
    CANCEL = 50  # close
    REJECT = 60  # close
    FULLY_FILLED = 70  # close
    CANCEL_REPLACE_TO_ZERO = 80  # close

    def __init__(self, subchain_id, open_event, open_reason, logger):
        """
        Subchain objects track the data around a subchain, including all the events, the open reason and the close
         reason.

        The open reasons are:
          * New Order
          * Cancel Replace to a different price
          * Cancel Replace that increases the open qty

        The close reasons are anything that opens a new subchain or closes out the entire order chain:
          * Cancel replace to a different price
          * Cancel Replace that increases qty
          * Cancel Report
          * Fully Filled

        In the event that a cancel replace changes price and increases quantity, the change in price should take
         precedent as to the reason for the open or close

        :param subchain_id:
        :param open_event:
        :param open_reason:
        """
        self._logger = logger
        self._events = []
        self._events.append(open_event)
        self._open_reason = open_reason
        self._close_reason = None
        self._close_event = None
        self._subchain_id = subchain_id
        self._logger.debug("OrderChain %s: Created new SubChain %s" % (str(self.chain_id()), str(self.subchain_id())))

    def add_event(self, event):
        if not self.is_open():
            raise Exception("Adding event to closed Subchain. EventID %s, Chain %s, SubChain %s" %
                            event.event_id(), event.chain_id(), self._subchain_id)
        self._logger.debug("OrderChain %s: Add %s %s to SubChain %s" %
                           (str(self.chain_id()), event.event_type_str(), str(event.event_id()),
                            str(self.subchain_id())))
        self._events.append(event)

    def close_subchain(self, close_reason):
        if self.is_open():
            self._close_event = close_reason
        else:
            raise Exception("Closing an already closed subchain: %s" % str(self.subchain_id()))

    def open_event(self):
        return self._events[0]

    def events(self):
        return self._events

    def is_open(self):
        return self._close_reason is None

    def chain_id(self):
        return self.open_event().chain_id()

    def subchain_id(self):
        return self._subchain_id


class OrderEventChain(object):
    def __init__(self, new_order_command, logger, subchain_id_generator):
        """
        OrderEventChain is data structure to track a complete chain of events that concern one order. This includes
         Commands, such as:
          * New Order
          * Cancel Replace
          * Cancel Request

        and Execution Reports, such as:
          * Acknowledgement
          * Reject
          * Partial Fill
          * Full Fill
          * Cancel Confirm

        Each Command should have at least one ExecutionReport and can have more.  Also, there can be Execution Reports
         that were not caused by a Command in the OrderEventChain, such as a fill that results from another order being
         the aggressor

        All Order Events in the OrderEventChain will share the same Chain ID.

        Further assumptions are made about all events in the same OrderEventChain:
          * For the same product
          * For the same side of the book
          * Events are in chronological order
          * Time in force (ex: FAK, FAR, or FOK) does not change
          * Order type (ex: market or limit) does not change

        An OrderEventChain also tracks subchains. Subchains are individual stretches of desired liquidity
         within an EventOrderChain. A new subchain is created with every user caused change in priority, assuming a
         standard order book. Changes in priority are triggered by:
          * New Order
          * Cancel Replace to a new price
          * Cancel Replace that increases the visible exposure (visible quantity), whether or not the price changes

        A subchain is closed by either the opening of a new subchain (a cancel replace to new price or up in visilbe
         quantity) or by the closing of the entire order chain (a Full Fill or a Cancel Confirm).

        Since matching engines do not typically track subchains or uniquely identify them, unique identifers have to
         be applied here. The passed in instance of IDGenerator will do this creation of IDs.

        :param new_order_command: MarketObjects.Events.OrderEvents.NewOrderCommand. NewOrderCommand that starts the chain.
        :param logger: logger.
        :param subchain_id_generator: the IDGenerator to be used for creating subchain_IDs
        """
        assert isinstance(new_order_command, NewOrderCommand)
        assert isinstance(subchain_id_generator, IDGenerator)
        self._logger = logger  # TODO add debug logging throughout

        # keeping this state local, rather than looking it up in self._new_order_command all the time speeds things up
        #  (only doing it for a select few that tend to get called more than others)
        self._chain_id = new_order_command.chain_id()
        self._user_id = new_order_command.user_id()
        self._side = new_order_command.side()

        self._new_order_command = new_order_command
        self._subchain_id_generator = subchain_id_generator

        # requested_exposures is a list in order of Exposure that has been requested by a Command but hasn't been
        #  acked, so isn't real exposure yet as it might not get acked. Will be None if no outstanding, unack'd request.
        self._requested_exposures = [Exposure(new_order_command.price(),
                                              new_order_command.qty(),
                                              new_order_command.event_id())
                                     ]

        # current_exposure is orderchain's currently open qty at price
        #  if no currently acknowledgments so no open exposure.
        self._current_exposure = None

        # add to list of events
        self._events = [new_order_command]
        self._filled_price_to_qty = defaultdict(int)
        self._logger.debug("New %s created, OrderChainID %s Market %s" %
                           (self.__class__.__name__, str(self.chain_id()), str(self.market())))
        # New Orders start new subchains, so start subchain and put it in list of subchains
        self._sub_chains = []
        self._visible_qty = 0  # an unack'd order has no qty showing on the book
        self._iceberg_peak_qty = new_order_command.iceberg_peak_qty()
        self._open = True
        self._event_id_to_cancel_replace_info = {}
        self._events_that_caused_visible_qty_refresh = set()
        self._price_at_close = None  # starts as none because this doesn't get populated until it is closed
        self._open_qty_at_close = None  # starts as none because isn't populated until it is closed
        self._match_ids = set()  # unique negotation ids that the order chain is part of

    def new_order_command(self):
        """
        Gets the NewOrderCommand that started the order chain

        :return: MarketObjects.Events.OrderEvents.NewOrderCommand
        """
        return self._new_order_command

    def chain_id(self):
        return self._chain_id

    def user_id(self):
        return self._user_id

    def visible_qty(self):
        # TODO document
        return self._visible_qty

    def hidden_qty(self):
        # TODO document
        # TODO unit test
        return self.current_qty() - self.visible_qty()

    def iceberg_peak_qty(self):
        # TODO document
        return self._iceberg_peak_qty

    def side(self):
        """
        Gets the Product of the OrderEventChain

        :return: MarketObjects.Side.Side
        """
        return self._side

    def is_limit_order(self):
        """
        Returns True if the order chain is for a limit order; false if it is not.

        :return: bool
        """
        return self.new_order_command().is_limit_order()

    def is_market_order(self):
        """
        Returns True if the new order is a market order; false if it is not.

        :return: bool
        """
        return self.new_order_command().is_market_order()

    def product(self):
        """
        Gets the Product of the OrderEventChain

        :return: MarketObjects.Product.Product
        """
        return self._new_order_command.market().product()

    def market(self):
        """
        Gets the Market of the OrderEventChain

        :return: MarketObjects.Market.Market
        """
        return self._new_order_command.market()

    def time_in_force(self):
        """
        Gets the Time in Force of the OrderEventChain

        :return: int
        """
        return self._new_order_command.time_in_force()

    def is_far(self):
        """
        True if the time in force of the chain is FAR
        
        :return: bool 
        """
        return self._new_order_command.is_far()

    def is_fak(self):
        """
        True if the time in force of the chain is FAK

        :return: bool 
        """
        return self._new_order_command.is_fak()

    def is_fok(self):
        """
        True if the time in force of the chain is FOK

        :return: bool 
        """
        return self._new_order_command.is_fok()

    def current_exposure(self):
        """
        Gets the current exposed (ie. acknowledged) pricelevel . If no exposure then returns None.

        :return: MarketObjects.PriceLevel.PriceLevel Could be None.
        """
        return self._current_exposure if self.current_exposure is not None else None

    def current_price(self):
        """
        Gets the currently acknowledged price.

        This is a helper function that is the same as self.current_exposure().price()

        :return: MarketObjects.Price.Price. Can be None
        """
        return self.current_exposure().price() if self.current_exposure() is not None else None

    def current_qty(self):
        """
        Gets the currently acknowledged qty.

        This is a helper function that is the same as self.current_exposure().qty(), returns 0 if current_exposure() is
         None

        :return: int. Can be None
        """
        return self.current_exposure().qty() if self.current_exposure() is not None else 0

    def most_recent_requested_exposure(self):
        """
        Gets the most recently requested exposure that has not been acknowledged. If no unacknowledged exposure requests
         then returns None
        :return: MarketObjects.PriceLevel.PriceLevel Could be None.
        """
        return None if len(self._requested_exposures) == 0 else self._requested_exposures[-1]

    def open_exposure_requests(self):
        return self._requested_exposures

    def events(self):
        """
        Gets the Event objects in the OrderChain in order of occurrence.

        :return: list of MarketObjects.Events.OrderEvents.OrderEvent
        """
        return self._events

    def most_recent_event(self):
        """
        Gets the most recent OrderEvent in the OrderEventChain.

        :return:  MarketObjects.Events.OrderEvents.OrderEvent
        """
        return self._events[-1]

    def last_update_time(self):
        """
        Gets the last time that the chain was updated (the time of the chain's last event).
        
        :return: float
        """
        return self._events[-1].timestamp()

    def most_recent_subchain(self):
        """
        Gets the most recent subchain in the OrderEventChain.

        Returns None if no subchains yet (which happens if no execution reports yet)

        :return:  MarketObjects.Events.EventChains.SubChain. Can be None.
        """
        return None if len(self._sub_chains) == 0 else self._sub_chains[-1]

    def subchains(self):
        """
        Returns the in-order list of subchains in the OrderEventChain.

        :return: list of MarketObjects.Events.EventChains.SubChain
        """
        return self._sub_chains

    def has_full_fill(self):
        """
        Determines if the orderchain contains a full fill

        :return: Bool, True if has FullFillReport, False otherwise
        """
        # since a full fill should be at end of event chain, more efficient to go backwards
        for event in reversed(self._events):
            if isinstance(event, FullFillReport):
                return True
        return False

    def match_ids(self):
        """
        Returns a set of the IDs of all the negotations that the order chain is
         part of.

        :return: set() of match ids
        """
        return self._match_ids

    def is_open(self):
        """
        Returns whether or not the order event chain is open. An event chain is open if it is still has exposure in the
         market. This means it has not received a Cancel Report, Full Fill Report, or has an any way had its qty set to
         0.
        :return: 
        """
        return self._open

    def has_partial_fill(self):
        """
        Determines if the orderchain contains a partial fill

        :return: Bool, True if has PartialFillReport, False otherwise
        """
        for event in self._events:
            if isinstance(event, PartialFillReport):
                return True
        return False

    def has_acknowledgement(self):
        """
        Determines if the orderchain contains an acknowledgement

        :return: Bool, True if has AcknowledgementReport, False otherwise
        """
        for event in self._events:
            if isinstance(event, AcknowledgementReport):
                return True
        return False

    def last_acknowledgement(self):
        """
        Gets the last acknowledgement in the order chain. Can be None.
        
        :return: Buttonwood.MarketObjects.Price.Price (can be None)
        """
        for event in reversed(self._events):
            if isinstance(event, AcknowledgementReport):
                return event
        return None

    def caused_visible_qty_refresh(self, event_id):
        """
        Is the event one that caused a refresh of the visible qty?
        
        :param event_id: unique identifier of event 
        :return: bool
        """
        return event_id in self._events_that_caused_visible_qty_refresh

    def find_requested_exposure(self, causing_event_id):
        """
        For a given event this returns the events requested exposure. 
        
        This can be None if the event did not result in a new requested exposure.
        
        :param causing_event_id: unique identifier of event. 
        :return: Exposure. (or None)
        """
        for exposure in self._requested_exposures:
            if exposure.causing_event_id() == causing_event_id:
                return exposure
        return None

    def cancel_replace_information(self, ack_event_id):
        """
        Cancel replace history kept for the ack of the cancel replace, so the
         event id passed in needs to be for an ack of a cancel replace. If it
         isn't then None will be returned.

        :param ack_event_id: unique identifier of the ack event
        :return: Bool. Can be None.
        """
        return self._event_id_to_cancel_replace_info.get(ack_event_id)

    def price_at_close(self):
        """
        The price of the chain at the time it was closed.

        Can be None for the following reasons:
         1) order event chain is not closed yet
         2) order was fully filled, so no exposure left, thus no price.
        """
        return self._price_at_close

    def open_qty_at_close(self):
        """
        The open qty of the chain at the time it was closed.

        Can be None for the following reasons:
         1) order event chain is not closed yet
         2) order was fully filled, so no exposure left, thus no price.
        """
        return self._open_qty_at_close

    # TODO track qty remaining at close

    def __str__(self):
        s = ""
        for event in self._events:
            event_json = event.to_json()
            s += json.dumps(event_json) + "\n"
        s += "\n%s: %s %s %s %d (%d) @ %s" % \
             (str(self.chain_id()), self.user_id(), str(self.market()), self.side(), self.visible_qty(),
              self.current_qty() - self.visible_qty(), str(self.current_price()))
        return s

    ############################################################################
    # Functions for updating the OrderEventChain  ##############################
    ############################################################################

    def _close_chain(self, closing_event):
        # TODO unit test!
        if self._current_exposure is not None and self._current_exposure.qty() > 0:
            self._price_at_close = self.current_exposure().price()
            self._open_qty_at_close = self.current_exposure().qty()
        else:  # for example, FAKs and other aggressive orders that get cancelled without an acknowledgement
            self._price_at_close = self.most_recent_requested_exposure().price()
            self._open_qty_at_close = self.most_recent_requested_exposure().qty()
        self._open = False
        self._visible_qty = 0
        # wipe out current exposure
        self._current_exposure = Exposure(None, 0, closing_event.event_id())
        # close out exposures
        self._requested_exposures = []

    def _close_requested_exposure(self, execution_report):
        """
        Close the exposure the execution_report is for

        :param execution_report:
        :return:
        """
        assert execution_report is not None, \
            "execution_report must be a valid ExecutionReport instance"
        assert isinstance(execution_report, ExecutionReport), \
            "execution_report must be a valid ExecutionReport instance"
        # partial fills can't close out exposure
        assert not isinstance(execution_report,
                              PartialFillReport), "PartialFillReports cannot close requested exposure."
        closed_exposure = False
        # if execution_report is a cancel confirm or full fill then close out all open requested exposures
        if isinstance(execution_report, CancelReport) or isinstance(execution_report, FullFillReport):
            self._requested_exposures = []  # TODO unit test this
            closed_exposure = True
        # if execution_report is an ack or a reject then close out the correct
        elif isinstance(execution_report, AcknowledgementReport) or isinstance(execution_report, RejectReport):
            response_to_event_id = execution_report.causing_command().event_id()
            # go through the open exposures and when/if we find the causing command event id, remove it
            index_to_remove = -1
            for i, requested_exposure in enumerate(self._requested_exposures):
                if requested_exposure.causing_event_id() == response_to_event_id:
                    # found it, so set the index for removal and since should only exist once, break the loop
                    index_to_remove = i
                    break
            del self._requested_exposures[index_to_remove]
            closed_exposure = True
        return closed_exposure

    def _close_current_subchain(self, subchain_close_reason):
        self.most_recent_subchain().close_subchain(subchain_close_reason)

    def _open_new_subchain(self, opening_cmd, open_reason):
        new_subchain = SubChain(self._subchain_id_generator.id(), opening_cmd, open_reason, self._logger)
        self._sub_chains.append(new_subchain)

    def apply_acknowledgement_report(self, ack):
        """
        Apply an acknowledgement to the OrderEventChain.

        :param ack: MarketObjects.Events.OrderEvents.AcknowledgementReport
        """
        assert isinstance(ack, AcknowledgementReport)
        assert ack.market() == self.market(), \
            "Acknowledgement does NOT have same market as the OrderEventChain expects"
        assert ack.chain_id() == self.chain_id(), \
            "Acknowledgement's chain ID (%s) does not match chain's ID (%s)" % (
            str(ack.chain_id()), str(self.chain_id()))
        assert ack.acknowledged_command().chain_id() == ack.chain_id(), \
            "Acknowledgements should have the same chain id as the command they are in response to."

        # append the ack to the underlying list of events
        self._events.append(ack)
        # close out the open exposure the ack is for
        if not self._close_requested_exposure(ack):
            raise Exception("Received an acknowledgement for event id %s but that event is not open in the chain." %
                            str(ack.response_to_command().event_id()))  # TODO unit test this behavior

        # if current exposure is not None then we need to set the cancel replace history
        ack_exposure = Exposure(ack.price(), ack.qty(), ack.event_id())
        if self.current_exposure() is not None:
            self._event_id_to_cancel_replace_info[ack.event_id()] = CancelReplaceInfo(self.current_exposure(),
                                                                                      ack_exposure, self.side(),
                                                                                      self.market())
        # handle subchains
        subchain_open_reason = None
        subchain_close_reason = None
        if isinstance(ack.acknowledged_command(), NewOrderCommand):
            subchain_open_reason = SubChain.NEW_ORDER
        elif isinstance(ack.acknowledged_command(), CancelReplaceCommand):
            if ack.qty() == 0:
                subchain_close_reason = SubChain.CANCEL_REPLACE_TO_ZERO
            else:
                # see if cancel replace to new price for current subchain close / new subchain open
                if self._event_id_to_cancel_replace_info[ack.event_id()].is_price_change():
                    subchain_close_reason = SubChain.CANCEL_REPLACE_PRICE
                    subchain_open_reason = SubChain.CANCEL_REPLACE_PRICE
                # see if cancel replace to more qty for current subchain close / new subchain open
                elif self._event_id_to_cancel_replace_info[ack.event_id()].is_qty_increase():
                    subchain_close_reason = SubChain.CANCEL_REPLACE_INCREASE_QTY
                    subchain_open_reason = SubChain.CANCEL_REPLACE_INCREASE_QTY
        # if we have a subchain reason then we need to close current subchain and open a new one
        # don't need to worry about opening and closing if this already happened due to a aggressing partial fill
        if self.most_recent_subchain() is None or self.most_recent_subchain().open_event().event_id() != ack.acknowledged_command().event_id():
            if subchain_close_reason is not None:
                self._close_current_subchain(subchain_close_reason)
            if subchain_open_reason is not None:  # if a new one is open then the ack'd event opens it
                self._open_new_subchain(ack.acknowledged_command(), subchain_open_reason)
            else:  # else the ack'd event gets added to already open SubChain
                self.most_recent_subchain().add_event(ack.acknowledged_command())
        # now, add the ack to the most recently open subchain
        self.most_recent_subchain().add_event(ack)

        # set the current exposure
        self._current_exposure = ack_exposure
        # set iceberg to new iceberg qty incase it changed
        self._iceberg_peak_qty = ack.iceberg_peak_qty()

        # visible qty gets resized if:
        # 1) new total qty is less than visible,
        # 2) a new price
        # 3) it is a cancel replace up in qty
        # 4) new visible is less than previous visible
        self._visible_qty = min(ack.iceberg_peak_qty(), ack.qty())

        # if the ack'd qty is 0, then this order chain is closed.
        if ack.qty() == 0:
            self._close_chain(ack)

    def apply_cancel_replace_command(self, cr):
        """
        Apply the cancel replace event to the OrderEventChain

        :param cr: MarketObjects.Events.OrderEvents.CancelReplaceCommand
        """
        assert isinstance(cr, CancelReplaceCommand)
        assert cr.market() == self.market(), \
            "Cancel Replace does NOT have same market as the OrderEventChain expects"
        assert cr.chain_id() == self.chain_id(), \
            "Cancel Replace's chain ID (%s) does not match chain's ID (%s)" % (str(cr.chain_id()), str(self.chain_id()))
        assert self.time_in_force() == OrderEventConstants.FAR, \
            "Cancel Replace commands only allowed for FAR time in force. This is a %s" % OrderEventConstants.time_in_force_str(
                self.time_in_force())
        self._events.append(cr)
        # set the requested exposure
        cr_exposure = Exposure(cr.price(), cr.qty(), cr.event_id())

        # add the requested exposure
        self._requested_exposures.append(cr_exposure)

    def apply_cancel_command(self, cc):
        """
        Apply the cancel command to the order chain

        :param cc: MarketObjects.Events.OrderEvents.CancelCommand
        """
        assert isinstance(cc, CancelCommand)
        assert cc.market() == self.market(), \
            "Cancel Command does NOT have same market as the OrderEventChain expects"
        assert cc.chain_id() == self.chain_id(), \
            "Cancel Commnad chain ID (%s) does not match chain's ID (%s)" % (str(cc.chain_id()), str(self.chain_id()))

        # append to list of events
        self._events.append(cc)
        # add a None exposure to the list of exposures
        self._requested_exposures.append(Exposure(None, 0, cc.event_id()))

    def apply_cancel_report(self, cr):
        """
        Apply the cancel report to the order chain.

        :param cr: MarketObjects.Events.OrderEvents.CancelReport
        """
        assert isinstance(cr, CancelReport)
        assert cr.market() == self.market(), \
            "Cancel Report does NOT have same market as the OrderEventChain expects"
        assert cr.chain_id() == self.chain_id(), \
            "Cancel Report chain ID (%s) does not match chain's ID (%s)" % (str(cr.chain_id()), str(self.chain_id()))
        # append to events
        self._events.append(cr)

        # if for some reason no subchain has been created to this point then one needs to be created and all events
        #  added to it. This could happen when for some reason a new order doesn't get acked, nor a cancel replace
        #  ack'd, and then the order is cancelled.
        #  This can also happen when there is a new order then cancel, like a FOK or FAK with no fill.
        if len(self._sub_chains) == 0:
            self._open_new_subchain(self._new_order_command, SubChain.NEW_ORDER)
            for event in self._events[1:]:  # skipping first one since already added in the creation of subchain
                self._sub_chains[0].add_event(event)
        else:
            # add to subchain
            if cr.cancel_command() is not None:  # need to add causing command to subchain
                self.most_recent_subchain().add_event(cr.cancel_command())

        self.most_recent_subchain().add_event(cr)
        # close open subchain
        self.most_recent_subchain().close_subchain(SubChain.CANCEL)
        # close order chain
        self._close_chain(cr)

    def _modify_exposure_by_partial_fill(self, pf):
        # if the aggressor then
        #  change outstanding exposure request because not ack'd yet
        #       - if outstanding exposure for causing event goes to zero or below then log an error
        #             and close order chain and subchain
        # don't need to change visible qty because not ack'd yet
        if pf.is_aggressor():
            requested_exposure = self.find_requested_exposure(pf.aggressing_command().event_id())
            if requested_exposure is None:
                self._logger.error(
                    "OrderChain state issue: %s. Aggressive partial fill (%s) with no outstanding exposure request for the aggressing event %s" %
                    (str(self.chain_id()), str(pf.event_id()), str(pf.aggressing_command().event_id())))
            else:
                requested_exposure.dec_qty(pf.fill_qty())
                if requested_exposure.qty() <= 0:
                    self._logger.error(
                        "OrderChain state issue: %s. Aggressive partial fill (%s) took requested exposure from %s to %d. Closing Order Chain." %
                        (str(self.chain_id()), str(pf.event_id()), str(pf.aggressing_command().event_id()),
                         pf.fill_qty()))
                    # close subchain down below after adding it
                    close_sub_chain = True
                    self._close_requested_exposure(pf)
                    self._close_chain(pf)
        else:
            # if passive then
            #  change open exposure by the fill size (because it has been ack'd)
            #       - if exposure goes to 0 or less then log a critical error and close order chain and subchain
            # change visible size
            #       - reduce by amount filled. If this is 0 or less then replenish min(iceberg peak qty, open exposure qty)
            if self._current_exposure is None:
                self._logger.error("OrderChain state issue: %s. Passive partial fill (%s) with no current exposure." %
                                   (str(self.chain_id()), str(pf.event_id())))
            else:
                if self._current_exposure.qty() - pf.fill_qty() <= 0:
                    self._logger.error(
                        "OrderChain state issue: %s. Passive partial fill (%s) took open exposure to %d" %
                        (str(self.chain_id()), str(pf.event_id()), pf.fill_qty()))
                    # subchain closed down below, after the partial fill is added to it
                    close_sub_chain = True
                    # don't decrement qty because this, in essence, gets taken care of in the close and we still want
                    #  some qty so we can grab the closing price
                    self._close_chain(pf)
                else:
                    self._current_exposure.dec_qty(pf.fill_qty())
                    self._visible_qty -= pf.fill_qty()
                    if self._visible_qty <= 0:
                        self._visible_qty = min(self.iceberg_peak_qty(), self._current_exposure.qty())
                        self._events_that_caused_visible_qty_refresh.add(pf.event_id())

    def apply_partial_fill_report(self, pf):
        """
        Apply the partial fill report to the order chain.

        :param pf: MarketObjects.Events.OrderEvents.PartialFillReport
        """
        assert isinstance(pf, PartialFillReport), "applyPartialFill called with an event that is not a PartialFillReport."
        # add to the list of events
        self._events.append(pf)
        # populate the map of qty at price for fills
        self._filled_price_to_qty[pf.fill_price()] = pf.fill_qty()
        # track the match id
        self._match_ids.add(pf.match_id())
        close_sub_chain = False
        self._modify_exposure_by_partial_fill(pf)
        if pf.is_aggressor():
            # if aggressor then could be opening a new subchain
            # but dont' need to worry about it if second partial fill of subchain, only handle it if aggressing event id doens't match the current subchain open id
            if self.most_recent_subchain() is None or self.most_recent_subchain().open_event().event_id() != pf.aggressing_command().event_id():
                subchain_open_reason = None
                subchain_close_reason = None
                if isinstance(pf.aggressing_command(), NewOrderCommand):
                    subchain_open_reason = SubChain.NEW_ORDER
                elif isinstance(pf.aggressing_command(), CancelReplaceCommand):
                    if self.most_recent_subchain() is not None:  # if it is then there is no subchain to close
                        subchain_close_reason = SubChain.CANCEL_REPLACE_PRICE
                    subchain_open_reason = SubChain.CANCEL_REPLACE_PRICE
                # if we have a subchain reason then we need to close current subchain and open a new one
                if subchain_close_reason is not None:
                    if self.most_recent_subchain() is not None:
                        self._close_current_subchain(subchain_close_reason)
                if subchain_open_reason is not None:  # if a new one is open then the ack'd event opens it
                    self._open_new_subchain(pf.aggressing_command(), subchain_open_reason)

        # add to the open subchain
        self.most_recent_subchain().add_event(pf)
        if close_sub_chain:
            self.most_recent_subchain().close_subchain(SubChain.FULLY_FILLED)

    def apply_full_fill_report(self, ff):
        """
        Apply the partial fill report to the order chain.

        :param ff: MarketObjects.Events.OrderEvents.FullFillReport
        """
        # TODO unit test
        assert isinstance(ff, FullFillReport), "applyFullFill called with an event that is not a FullFillReport."
        # add to the list of events
        self._events.append(ff)
        # populate the map of qty at price for fills
        self._filled_price_to_qty[ff.fill_price()] = ff.fill_qty()
        # track the match id
        self._match_ids.add(ff.match_id())
        # log warning if amount filled wouldn't actually fully fill the chain (using open requested exposure if aggressor and acked exposure if passive)
        if ff.is_aggressor():
            requested_exposure = self.find_requested_exposure(ff.aggressing_command().event_id())
            if ff.fill_qty() > requested_exposure.qty():
                self._logger.warn("OrderChain state issue: %s. Aggressive full fill (%s) for %d is more than open exposure for %s. Took open exposure to %d" %
                                  (str(self.chain_id()), str(ff.event_id()), ff.fill_qty(),
                                   str(requested_exposure.causing_event_id()),
                                   requested_exposure.qty() - ff.fill_qty()))
            elif ff.fill_qty() < requested_exposure.qty():
                self._logger.warn("OrderChain state issue: %s. Aggressive full fill (%s) for %d is less than open exposure for %s. Took open exposure to %d. Should NOT be full fill." %
                                  (str(self.chain_id()), str(ff.event_id()), ff.fill_qty(),
                                   str(requested_exposure.causing_event_id()),
                                   requested_exposure.qty() - ff.fill_qty()))
            # if subchain for the aggressor command is not already open then need to open it and close previous
            if self.most_recent_subchain() is None or self.most_recent_subchain().open_event().event_id() != ff.aggressing_command().event_id():
                # cancel replace price is only thing that would result in a new subchain from a full fill
                if self.most_recent_subchain() is not None:
                    self.most_recent_subchain().close_subchain(SubChain.CANCEL_REPLACE_PRICE)
                self._sub_chains.append(
                    SubChain(self._subchain_id_generator.id(), ff.aggressing_command(), SubChain.CANCEL_REPLACE_PRICE,
                             self._logger))
        else:
            if ff.fill_qty() > self._current_exposure.qty():
                self._logger.warn("OrderChain state issue: %s. Passive full fill (%s) for %d is more than current exposure. Took exposure to %d" %
                                  (str(self.chain_id()), str(ff.event_id()), ff.fill_qty(),
                                   self._current_exposure.qty() - ff.fill_qty()))
            elif ff.fill_qty() < self._current_exposure.qty():
                self._logger.warn("OrderChain state issue: %s. Passive full fill (%s) for %d is less than current exposure. Took exposure to %d. Should NOT be full fill." %
                                  (str(self.chain_id()), str(ff.event_id()), ff.fill_qty(),
                                   self._current_exposure.qty() - ff.fill_qty()))
        # add to the open subchain
        self.most_recent_subchain().add_event(ff)
        # close the open subchain
        self.most_recent_subchain().close_subchain(SubChain.FULLY_FILLED)
        # close the order chain
        self._close_chain(ff)

    def apply_reject_report(self, rej):
        """
        Apply a reject to the OrderEventChain.

        :param rej: MarketObjects.Events.OrderEvents.RejectReport
        """
        assert isinstance(rej, RejectReport)
        assert rej.market() == self.market(), \
            "Reject does NOT have same market as the OrderEventChain expects"
        assert rej.chain_id() == self.chain_id(), \
            "Reject's chain ID (%s) does not match chain's ID (%s)" % (str(rej.chain_id()), str(self.chain_id()))
        assert rej.rejected_command().chain_id() == rej.chain_id(), \
            "Reject should have the same chain id as the command they are in response to."
        # add reject and what is being rejected to the subchain.
        self._events.append(rej)
        # add to subchain
        if self.most_recent_subchain() is not None:
            self.most_recent_subchain().add_event(rej.rejected_command())
        else:  # else it si the new order that needs to be added (new order is the only way we get here)
            self._sub_chains.append(
                SubChain(self._subchain_id_generator.id(), rej.rejected_command(), SubChain.NEW_ORDER, self._logger))
        self.most_recent_subchain().add_event(rej)

        # close out the exposure that is open
        self._close_requested_exposure(rej)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from cdecimal import Decimal
from buttonwood.MarketObjects.Side import Side


class Price(object):
    def __init__(self, price_value):
        """
        Represents a price. Is created with a string, for accuracy, or as a Decimal and is kept as a Decimal
  
        :param price_value: str
        """
        assert isinstance(price_value, str) or isinstance(price_value, Decimal)
        self._price = Decimal(price_value)  # yup, this is do-able. You can instantiate a Decimal with another Decimal
        self._hash = hash(price_value)

    def price(self):
        """
        Gets the underlying price of the Price object as a Decimal
        :return: Decimal
        """
        return self._price

    def better_than(self, other_price, side):
        """
        Determines if this price is better than the passed in price. Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then better than means a higher price
           *if side is ask then better than means a lower price
           *if side is other then there is no such thing as a better price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        assert isinstance(other_price, Price), "other_price should be MarketObjects.Price.Price"
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        # if this price is a bid it is better than otherPrice if greater than

        if side.is_bid():
            return self._price > other_price._price
        # otherwise price is an ask and it is better than otherPrice if less than
        else:
            return self._price < other_price._price

    def better_or_same_as(self, other_price, side):
        """
        Determines if this price is better than or same as the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then better than means a higher price
           *if side is ask then better than means a lower price
           *if side is other then there is no such thing as a better price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        assert isinstance(other_price, Price), "other_price should be MarketObjects.Price.Price"
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        if other_price._price == self._price:
            return True
        return self.better_than(other_price, side)

    def worse_than(self, other_price, side):
        """
        Determines if this price is worse than the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then worse than means a lower price
           *if side is ask then worse than means a higher price
           *if side is other then there is no such thing as a worse price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        # if this price is a bid it is worse than otherPrice if less than
        assert isinstance(other_price, Price), "other_price should be MarketObjects.Price.Price"
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        if side.is_bid():
            return self._price < other_price._price
        # otherwise price is an ask and it is worse than otherPrice if greater than
        else:
            return self._price > other_price._price

    def worse_or_same_as(self, other_price, side):
        """
        Determines if this price is worse than or same as the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then worse than means a lower price
           *if side is ask then worse than means a higher price
           *if side is other then there is no such thing as a worse price so the return is False

        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        assert isinstance(other_price, Price), "other_price should be MarketObjects.Price.Price"
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        if other_price._price == self._price:
            return True
        return self.worse_than(other_price, side)

    def ticks_behind(self, other_price, side, product):
        """
        For bids it subtracts its own price from the other price and divides by mpi
  
        For asks it subtracts the other price from its own price and divides by mpi.
  
        The result is if the other price is a better price then we get a positive
         number and if the other price is a worse price, we get a negative number.
  
        If otherprice is none it returns None
        :param other_price: MarketObjects.Price.Price the price you are comparing too
        :param side: MMarketObjects.Side.Side side of the market the comparison is happening on
        :param product: MarketObjects.Product.Product
        :return: float Can be None
        """
        if other_price is None:
            return None

        if side.is_bid():
            return (other_price._price - self._price) / product.mpi()
        else:
            return (self._price - other_price._price) / product.mpi()

    def __lt__(self, other):
        if isinstance(other, Price):
            return self._price < other._price
        if isinstance(other, (float, int, Decimal)):
            return self._price < other
        return None

    def __le__(self, other):
        if isinstance(other, Price):
            return self._price <= other._price
        if isinstance(other, (float, int, Decimal)):
            return self._price <= other
        return None

    def __gt__(self, other):
        if isinstance(other, Price):
            return self._price > other._price
        if isinstance(other, (float, int, Decimal)):
            return self._price > other
        return None

    def __ge__(self, other):
        if isinstance(other, Price):
            return self._price >= other._price
        if isinstance(other, (float, int, Decimal)):
            return self._price >= other
        return None

    def __eq__(self, other):
        if isinstance(other, Price):
            return self._price == other._price
        if isinstance(other, (float, int, Decimal)):
            return self._price == other
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(self.price())

    def __rsub__(self, other):
        return other - self._price

    def __sub__(self, other):
        return self._price - other

    def __radd__(self, other):
        return other + self._price

    def __add__(self, other):
        return self._price + other

    def __div__(self, other):
        return self._price / self.__math_val(other)

    def __mul__(self, other):
        return self._price * self.__math_val(other)

    def __math_val(self, obj):
        val = obj
        if isinstance(obj, Price):
            val = obj.price()
        return val



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.utils import timehelpers


class TimeRange(object):
    def __init__(self, start_time, payload):
        self._start_time = start_time
        self._end_time = None
        self._payload = payload

    def start_time(self):
        return self._start_time

    def end_time(self):
        return self._end_time

    def set_end_time(self, time):
        if time < self._start_time:
            raise Exception("%.6f is less than start time: %.6f" % (self._start_time, time))
        self._end_time = time

    def is_closed(self):
        return self._end_time is not None

    def payload(self):
        return self._payload

    def __str__(self):
        return "%.6f - %.6f : %s" % (self.start_time(), self.end_time(), self.payload())


class TimeRangeList(object):
    def __init__(self):
        self._list = []

    def __len__(self):
        return len(self._list)

    def set_last_end_time(self, end_time):
        self._list[-1].set_end_time(end_time)

    def last_payload(self):
        return self._list[-1].payload()

    def append(self, time_range_item):
        assert isinstance(time_range_item, TimeRange)
        assert len(self._list) == 0 or (not self._list[-1].is_closed()) or (time_range_item.start_time() >= self._list[-1].end_time()), "New item's start time must be >= last item's end time"
        if len(self._list) > 0 and not self._list[-1].is_closed():
            self.set_last_end_time(time_range_item.start_time())
        self._list.append(time_range_item)

    def items_in_window(self, window_start_time, window_end_time, enforce_window=True):
        items = []
        for time_range_item in self._list:
            # we care about three options.
            # 1) start_time is between time_range_item start and end, or
            # 2) end_time is between time_range_itme start and end
            # 3) the entire time_range_item falls within the window
            # if the time_range falls within the entire window, you use it and only it, whether enforcing the window or not
            tr_start_time = time_range_item.start_time()
            tr_end_time = time_range_item.end_time()
            if tr_start_time >= window_start_time and time_range_item.is_closed() and tr_end_time <= window_end_time:
                items.append(time_range_item)
            # otherwise, if there is any window overlap then you do the max and mins
            elif time_range_item.is_closed() and \
                    ((time_range_item.start_time() <= window_start_time <= time_range_item.end_time()) or
                     (time_range_item.start_time() <= window_end_time <= time_range_item.end_time())):
                if enforce_window:
                    use_start_time = max(window_start_time, time_range_item.start_time())
                    use_end_time = min(window_end_time, time_range_item.end_time())
                    tr = TimeRange(use_start_time, time_range_item.payload())
                    tr.set_end_time(use_end_time)
                    items.append(tr)
                else:
                    items.append(time_range_item)
            elif not time_range_item.is_closed() and tr_start_time < window_end_time:
                if enforce_window:
                    use_start_time = max(window_start_time, time_range_item.start_time())
                    use_end_time = window_end_time
                    tr = TimeRange(use_start_time, time_range_item.payload())
                    tr.set_end_time(use_end_time)
                    items.append(tr)
                else:
                    items.append(time_range_item)
            # TODO should make sure that anything else is ust outside of the window and closed
        return items

    def trim_to_start_time(self, start_time):
        trim_to = 0
        for index, time_range_item in enumerate(self._list):
            if time_range_item.is_closed() and time_range_item.end_time() < start_time:
                trim_to = index + 1
            else:
                break
        if trim_to > 0:  # only bother doing it if no change
            if trim_to >= len(self._list):
                self._list = self._list[len(self._list) - 1:]  # this way we never delete the last item (# TODO  this might be redundant safety)
            else:
                self._list = self._list[trim_to:]
            if len(self._list) == 0:
                raise Exception("List ended up at 0 length! Something not right! (%s)" % str(timehelpers.epoch_to_timestamp(start_time)))  # TODO delete this, should never happen



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json


class BasicEvent(object):

    def __init__(self, event_id, timestamp):
        """
        Event is a fairly generic object. It is the base object for any market or market impacting event. There is not
        much that is shared by all events, resulting in a bare bones object.

        :param event_id: the unique identifier of the event
        :param timestamp: float. The time of the event in standard python representation of microseconds where left of 
                            the decimal is the number of seconds and to the right of the decimal is the number of us.
        """
        assert isinstance(timestamp, float)
        self._event_id = event_id
        self._timestamp = timestamp

    def event_type_str(self):
        raise NotImplemented("event_type_str to be implemented by inheriting sub class.")

    def timestamp(self):
        """
        Get the timestamp of an event in standard python representation of microseconds where left of the
         decimal is the number of seconds and to the right of the decimal is the number of microseconds.

        :return: float
        """
        return self._timestamp

    def event_id(self):
        """
        Get the unique identifier of the event.

        :return: the identifier
        """
        return self._event_id

    def to_json(self):
        return {self.__class__.__name__: {"event_id": self.event_id(), "timestamp": self.timestamp()}}

    def __str__(self):
        return json.dumps(self.to_json())



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import random


class IDGenerator:

    def __init__(self):
        pass

    def id(self):
        """
        Return the next ID.

        :return: the next ID
        """
        raise NotImplemented("id() to be implemented by the inheriting class.")

    def last_id(self):
        """
        Gets the ID that was returned by the last id() call.

        :return: the last id
        """
        raise NotImplemented("last_id() to be implemented by the inheriting class.")


class MonotonicIntID(IDGenerator):

    def __init__(self, seed=0, increment=1, max_id=sys.maxint):
        """
        A monotonically increasing integer.  increases by the passed the increment value, starting at the seed value.
         The seed value is reserved. It itself will not be returned as an ID

        :param seed:
        :param increment:
        """
        assert isinstance(seed, int) or isinstance(seed, long)
        assert isinstance(increment, int)
        IDGenerator.__init__(self)
        self._seed = seed
        self._increment = increment
        self._value = seed
        self._max_id = max_id

    def id(self):
        self._value += self._increment
        if self._value > self._max_id:
            raise Exception("ID would be greater than %d, which generator cannot handle" % self._max_id)
        return self._value

    def last_id(self):
        return self._value

    def seed(self):
        return self._seed

    def increment(self):
        return self._increment


class RandomPositiveIntID(IDGenerator):

    def __init__(self, floor=1, ceiling=sys.maxint, attempts=20000):
        assert isinstance(floor, int)
        assert isinstance(ceiling, int)
        assert ceiling > floor
        IDGenerator.__init__(self)
        self._floor = floor
        self._ceiling = ceiling
        self._attempts = attempts
        self._ids = set()
        self._id_count = 0 # should be cheaper to do this then do len over and over again
        self._max_id_count = ceiling - floor + 1

    def id(self):
        if self._id_count >= self._max_id_count:
            raise Exception("%d IDs distributed. All possible IDs have been used." % self._id_count)
        for x in xrange(1, self._attempts):
            x = random.randint(self._floor, self._ceiling)
            if x not in self._ids:
                self._ids.add(x)
                self._id_count += 1
                return x
        raise Exception("%d attempts at generating an ID. You need a bigger ID universe or to allow for more attempts" % self._attempts)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.Side import ASK_SIDE
from buttonwood.MarketObjects.Side import BID_SIDE


class BasicOrderBook(object):

    """
    BasicOrderBook includes the base functions for querying an order book,
     no matter the type of order book (price level, order level, or
     otherwise).

    As many of the functions are implemented as can be, and those that aren't
     (as indicate by raising a NotImplemented exception) need to be implemented
     by the inhereting class.

    The inhereting class is also responsible for creating and maintaining the
     underlying data structures as well as the functions (internal and exposed)
     that allow for the manipulation of that structure.

    :param market: MarketObjects.Market.Market
    :param logger: logger
    """

    def __init__(self, market, logger):
        self._logger = logger
        self._market = market

    def market(self):
        """
        Get the market of the order book.

        :return: MarketObjects.Market.Market
        """
        return self._market

    def name(self):
        """
        The name of the order book to be used in logging and other
         identification.

        :return: str
        """
        raise NotImplemented("%s: name() not implemented." % self.__class__.__name__)

    def last_update_time(self):
        """
        Get the last "market time" the order book was updated.

        Order books should be thought of as living breathing things that update
         when new information is applied, so knowing when the last update
         occurred is important.

        "Market Time" vs "clock time".
         Market Time is the time of the market, which is time that comes from
          the market messages.
         Clock time is the time of the server if one called a system call to get
          the current time.
         The last update time is market time and *not* clock time. Using Market
          Time throughout allows for analytics to be done on historical data
          and on live data removes the issue of transport time and queues
          throwing off the deltas between updates.

        The time that gets returned is microseconds since the epoch in seconds
         notation, where to the left of the decimal is an integer and to the
         right of the decimal is the microseconds

        :return: float
        """
        raise NotImplemented("%s: last_update_time() not implemented." % self.__class__.__name__)

    def price_is_bid(self, price):
        """
        Is the price a bid or not. True if the price is worse than or equal to
         the top of book bid price. If no top of book bid price then return
         False.

        :return: boolean
        """
        tob_bid = self.best_bid_price()
        if tob_bid is None:
            return False
        return price.worse_or_same_as(tob_bid, BID_SIDE)

    def price_is_ask(self, price):
        """
        Is the price an ask or not. True if the price is worse than or equal to
         the top of book ask price. If no top of book ask price then return
         False.

        :return: boolean
        """
        tob_ask = self.best_ask_price()
        if tob_ask is None:
            return False
        return price.worse_or_same_as(tob_ask, ASK_SIDE)

    def bid_prices(self):
        """
        All the bid prices currently live in the book in order from best to
         worst (highest to lowest).

        :return: list of MarketObjects.Price.Price
        """
        return self.prices(BID_SIDE)

    def ask_prices(self):
        """
        All the bid prices currently live in the book in order from best to
         worst (lowest to highest).

        :return: list of MarketObjects.Price.Price
        """
        return self.prices(ASK_SIDE)

    def prices(self, side):
        """
        All the prices currently live in the book at the specified side in order
         from best to worst price.

        :param side: MarketObjects.Side.Side
        :return: list of MarketObjects.Price.Price from best to worst
        """
        raise NotImplemented("%s: prices(side) not implemented." % self.__class__.__name__)

    def best_bid_price(self):
        """
        Returns the best price of the bid side of the book.
        Will return None if the side is empty.

        :return: MarketObjects.Price.Price. Can be None.
        """
        return self.best_price(BID_SIDE)

    def best_ask_price(self):
        """
        Returns the best price of the ask side of the book.
        Will return None if the side is empty.

        :return: MarketObjects.Price.Price. Can be None.
        """
        return self.best_price(ASK_SIDE)

    def best_price(self, side):
        """
        Returns the best price of the of the book (top of book price) for the
         specified side. Will return None if the side is empty

        :param: MarketObjects.Side.Side. Side getting TOB for.
        :return: MarketObjects.Price.Price.
        """
        prices = self.prices(side)
        if len(prices) == 0:
            return None
        return prices[0]

    def best_bid_level(self):
        """
        Gets the top of book price level on the bid

        :return: MarketObjects.PriceLevel.PriceLevel
        """
        return self.best_level(BID_SIDE)

    def best_ask_level(self):
        """
        Gets the top of book price level on the bid

        :return: MarketObjects.PriceLevel.PriceLevel
        """
        return self.best_level(ASK_SIDE)

    def best_level(self, side):
        """
        Gets the top of book price level for the given side

        :return: MarketObjects.PriceLevel.PriceLevel
        """
        raise NotImplemented("%s: best_level(side) not implemented." % self.__class__.__name__)

    def total_qty_at_price(self, side, price):
        """
        Gets the total quantity (which is visible + hidden) at a price for the
         given side.

        :param side: MarketObjects.Side.Side
        :param price: MarketObjects.Price.Price
        :return: int
        """
        return self.visible_qty_at_price(side, price) + self.hidden_qty_at_price(side, price)

    def visible_qty_at_price(self, side, price):
        """
        Gets the visible quantity at the price for the given side.

        :param side: MarketObjects.Side.Side
        :param price: MarketObjects.Price.Price
        :return: int
        """
        raise NotImplemented("%s visible_qty_at_price(side, price) not implemented" % self.__class__.__name__)

    def hidden_qty_at_price(self, side, price):
        """
        Gets the hidden quantity at the price for the given side.

        :param side: MarketObjects.Side.Side
        :param price: MarketObjects.Price.Price
        :return: int
        """
        raise NotImplemented("%s hidden_qty_at_price(side, price) not implemented" % self.__class__.__name__)

    def num_orders_at_price(self, side, price):
        """
        Get the number of orders at a price for the given side.

        :param side: MarketObjects.Side.Side.
        :param price: MarketObjects.Price.Price.
        :return: int.
        """
        raise NotImplemented("%s num_orders_at_price(side, price, include_hideen) not implemented" % self.__class__.__name__)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.Side import BID_SIDE, ASK_SIDE
from buttonwood.utils.dicts import NDeepDict
from collections import defaultdict


class MarketOrderTicksFromCrossing(OrderLevelBookListener, OrderEventListener):
    """
    Tracks how far away a MarketOrder was from crossing the opposite TOB.
     * Greater than 0 is how many ticks away from crossing (but did not cross)
     * 0 means it crossed TOB
     * less than 0 is how many ticks through the opposite TOB the order went
     * None means there was no opposite side to cross (or the order chain id
       being asked about wasn't a market order)

    Tracks only for order chain id because at the time of the new order, the
     subchain hasn't been created yet (doesn't get created until the execution
     report) so does not have a subchain id to use.

    This is designed so it can work with mulitiple order books at once.
    """
    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._market_chain_id_ticks = NDeepDict(2)
        self._market_side_tob = {}
        self._market_side_tob[BID_SIDE] = defaultdict()
        self._market_side_tob[ASK_SIDE] = defaultdict()

    def handle_new_order_command(self, new_order_command, resulting_order_chain):
        # only applies to new order commands
        if new_order_command.is_market_order():
            side = new_order_command.side()
            price = new_order_command.price()
            market = new_order_command.market()
            mpi = market.product().mpi()
            opp_price = self._market_side_tob[market][side.other_side()]
            ticks_away = ((opp_price - price) if side.is_bid() else (price - opp_price)) / mpi
            self._market_chain_id_ticks.set([market, new_order_command.chain_id()], value=ticks_away)

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        Every time an orderbook comes in and tob updated need to save the bid price and the ask
         price. And only need to update for the side that is being updated.
        """
        if tob_updated:
            market = order_book.market()
            side = causing_order_chain.side()
            self._market_side_tob[market][side] = order_book.best_price(side)

    def ticks_from_crossing(self, market, chain_id):
        """
        Gets how many ticks away from crossing the chain id was.
         * Greater than 0 is how many ticks away from crossing (but did not cross)
         * 0 means it crossed TOB
         * less than 0 is how many ticks through the opposite TOB the order went
         * None means there was no opposite side to cross (or the order chain id
           being asked about wasn't a market order)

        :param market: MarketObjects.Market.Market
        :param chain_id: order chain's unique identifier
        :return: int
        """
        return self._market_chain_id_ticks.get([market, chain_id])



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.PriceLevel import PriceLevel
from buttonwood.MarketObjects.Side import Side
from buttonwood.MarketObjects.Market import Market
from buttonwood.MarketObjects.Price import Price


class Quote(object):
    def __init__(self, market, side, price, visible_qty, hidden_qty=0):
        """
        An individual quote expressing a desire to buy or sell a Product for a Price. Visible qty is required and
         hidden qty is optional, defaulting to 0 because many markets don't support hidden quantity.
  
        :param market: MarketObjects.Market.Market.
        :param side: Side.
        :param price: Price.
        :param visible_qty: int. Must be greater than 0
        :param hidden_qty: int. Must be 0 or more. Optional. Defaults to 0.
        """
        assert isinstance(price, Price) or isinstance(price, str)
        assert isinstance(side, Side)
        assert isinstance(market, Market)
        assert isinstance(visible_qty, int)
        assert isinstance(hidden_qty, int)
        assert visible_qty > 0, "A quote's visible qty must be greater than 0"
        assert hidden_qty >= 0, "A quote's hidden qty must be greater than or equal to 0"
        use_price = price
        if isinstance(price, str):
            use_price = Price(price)
        assert market.product().is_valid_price(use_price), \
            "%s is not a valid price for a product with minimum price increment %s" % \
            (str(use_price), str(market.product().mpi()))
        self._side = side
        self._market = market
        self._price_level = PriceLevel(use_price, visible_qty, hidden_qty, num_orders=1)

    def side(self):
        """
        Get the side of the market the quote is for
  
        :return: Side
        """
        return self._side

    def market(self):
        """
        Get the market the quote is for
  
        :return: MarketObjects.Market.Market
        """
        return self._market

    def price(self):
        """
        Get the price of the quote
  
        :return: Price.
        """
        return self._price_level.price()

    def visible_qty(self):
        """
        Get the visible qty of the quote
  
        :return: int
        """
        return self._price_level.visible_qty()

    def hidden_qty(self):
        """
        Get the hidden qty of the quote
  
        :return: int
        """
        return self._price_level.hidden_qty()

    def total_qty(self):
        """
        Get the total qty of the quote.
  
        :return: int.
        """
        return self._price_level.hidden_qty() + self._price_level.visible_qty()

    def is_buy(self):
        """
        Tells if you the Quote is a buy or not
  
        :return: bool
        """
        return self._side.is_bid()

    def is_sell(self):
        """
        Tells if you the Quote is a sell or not
  
        :return: bool
        """
        return self._side.is_ask()

    def __eq__(self, other):
        if not isinstance(other, Quote):
            return False
        return self._price_level == other._price_level and \
               self.side() == other.side() and \
               self.market() == other.market()

    def __str__(self):
        return "%s Quote: %s %d (%d visible, %d hidden) at %s" % \
               (str(self.market()),
                "Buy" if self.is_buy() else "Sell",
                self.total_qty(),
                self.visible_qty(),
                self.hidden_qty(),
                str(self.price()))



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.Side import BID_SIDE
from buttonwood.MarketObjects.Side import ASK_SIDE
from buttonwood.MarketObjects.Events.OrderEvents import FillReport


LAST_TRADE_TYPE = 0
BID_TOB_TYPE = 1
ASK_TOB_TYPE = 2
MID_TYPE = 3

def bid_price(order_book, causing_order_chain):
    price = None
    if order_book is not None:
        price = order_book.best_price(BID_SIDE)
    return price


def ask_price(order_book, causing_order_chain):
    price = None
    if order_book is not None:
        price = order_book.best_price(ASK_SIDE)
    return price


def mid_price(order_book, causing_order_chain):
    price = None
    if order_book is not None:
        best_ask_price = order_book.best_price(ASK_SIDE)
        best_bid_price = order_book.best_price(BID_SIDE)
        if best_ask_price is not None and best_bid_price is not None:
            price = (best_bid_price + best_ask_price) / 2.0
        elif best_ask_price is not None:
            # ask price is not None but bid price is, use ask
            price = best_ask_price
        elif best_bid_price is not None:
            # bid price is not None but ask price is, use bid
            price = best_bid_price
        else:
            # both sides are none so no price
            price = None
    return price


def last_trade(order_book, causing_order_chain):
    price = None
    if isinstance(causing_order_chain.most_recent_event(), FillReport):
        price = causing_order_chain.most_recent_event().fill_price()
    return price


class OHLCPriceTracker(object):

    def __init__(self, start_time, start_price):
        self._start_time = start_time
        self._open = start_price
        self._close = start_price
        self._high = start_price
        self._low = start_price

    def update_price(self, price):
        # price cannot be higher than the high and lower than the low so elif works here
        if price > self._high:
            self._high = price
        elif price < self._low:
            self._low = price
        # always update close price as this might be last price of period
        self._close = price

    def start_time(self):
        return self._start_time

    def open(self):
        return self._open

    def close(self):
        return self._close

    def high(self):
        return self._high

    def low(self):
        return self._low


class ConfigurablePeriodBarListener(OrderLevelBookListener):
    """
    Creates Period Bars.

    Takes period, start time and stop time as arguments.

    Period is number of seconds with fractions of seconds being part of the decimal.

    Start time and stop time are both optional number of seconds since the epoch.

    If start time is None then will use first time we see an update, as long
      as that time is one full period less than the stop time.

    If stop time is None it is considered infinite and will go on as long as there is data.
        
    """

    TYPE_TO_FUNCTION = {LAST_TRADE_TYPE: last_trade,
                        BID_TOB_TYPE: bid_price,
                        ASK_TOB_TYPE: ask_price,
                        MID_TYPE: mid_price}

    # TODO UNIT TEST
    def __init__(self, logger, period, start_time=None, stop_time=None, type = LAST_TRADE_TYPE):
        OrderLevelBookListener.__init__(self, logger)
        assert start_time is None or (stop_time is None or start_time <= stop_time - period), "If start time is None and stop time is not None, then start time must be less than or equal to the stop time minus one period."
        self._start_time = start_time
        self._previous_time = start_time
        self._stop_time = stop_time
        self._period = period
        self._market_to_bars = {}
        self._market_to_volume = {}
        self._market_to_current_tracker = {}
        self._type = type
        self._price_func = self.TYPE_TO_FUNCTION[type]
        self._next_stop_time = start_time + period

    def type(self):
        return self._type

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        if tob_updated:
            time = order_book.last_update_time()
            if time >= self._start_time:
                market = order_book.market()
                if market not in self._market_to_bars:
                    self._market_to_bars[market] = []

                # if time is < previous time we have an issue.
                if time < self._previous_time:
                    raise Exception("Book update at %0.6f is before the time of the previous update at %0.6f" % (time, self._previous_time))

                # don't do anything if before start time
                if time < self._start_time:
                    return

                if


        tob = self._market_to_previous_tob.get(market)
        event_id = causing_order_chain.most_recent_event().event_id()
        self._event_id_to_tob[event_id] = tob
        # set the previous tob to the new tob (but only do it if TOB changed
        if tob_updated:
            self._market_to_previous_tob[market] = (order_book.best_level(BID_SIDE), order_book.best_level(ASK_SIDE))

    def clean_up_order_chain(self, order_chain):
        for event in order_chain.events():
            del self._event_id_to_tob[event.event_id()]

    def tob_before_event(self, event_id):
        """
        Gets the top of book snapshots as a tuple (bid side PriceLevel, ask side PriceLevel) for the given event_id.
        
        Can be None for either side if the side does not exist in the order book before the event_id.
        
        Can be None instead of a tuple if the event_id is not known.
        
        :param event_id: unique identifier of the event 
        :return: (Buttonwood.MarketObjects.PriceLevel.PriceLevel, Buttonwood.MarketObjects.PriceLevel.PriceLevel)
        """
        return self._event_id_to_tob.get(event_id)


class TopOfBookAfterEventListener(OrderLevelBookListener):
    """
    Tracks top of book after an event for the market the event occurred in. Keeps track in dict of 
      event_id -> (PriceLevel, PriceLevel), where the first PriceLevel is the bid and second is the ask

      if event does not impact the book it doesn't end up getting tracked.
    """

    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._event_id_to_tob = {}

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        event_id = causing_order_chain.most_recent_event().event_id()
        # print type(causing_order_chain.most_recent_event())
        self._event_id_to_tob[event_id] = (order_book.best_level(BID_SIDE), order_book.best_level(ASK_SIDE))

    def clean_up_order_chain(self, order_chain):
        for event in order_chain.events():
            # not all events cause notify_book_update to be called
            if event.event_id() in self._event_id_to_tob:
                del self._event_id_to_tob[event.event_id()]

    def tob_after_event(self, event_id):
        """
        Gets the top of book snapshots as a tuple (bid side PriceLevel, ask side PriceLevel) for the given event_id.

        Can be None for either side if the side does not exist in the order book after the event_id.

        Can be None instead of a tuple if the event_id is not known.

        :param event_id: unique identifier of the event 
        :return: (Buttonwood.MarketObjects.PriceLevel.PriceLevel, Buttonwood.MarketObjects.PriceLevel.PriceLevel)
        """
        return self._event_id_to_tob.get(event_id)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.Quote import Quote


class TwoSidedQuote(object):

    def __init__(self, buy_quote, sell_quote, allow_cross=False):
        """
        This is a basic wrapper for a Quote, that has a bid and offer quote.

        :param buy_quote: the quote to buy
        :param sell_quote: the quote to sell
        :param allow_cross: bool. Whether the market can be crossed/locked or not. Optional. Defaults to False (not allowed)
        """
        assert buy_quote is None or isinstance(buy_quote, Quote)
        assert sell_quote is None or isinstance(sell_quote, Quote)
        assert buy_quote is None or buy_quote.side().is_bid(), "If buy quote is not None it's Side must be a bid."
        assert sell_quote is None or sell_quote.side().is_ask(), "If sell quote is not None it's Side must be an ask."
        assert isinstance(allow_cross, bool)

        self._buy_quote = buy_quote
        self._sell_quote = sell_quote
        self._allow_cross = allow_cross
        self._validate_quotes()

    def _validate_quotes(self):
        if self._buy_quote is None or self._sell_quote is None:
            return

        if not self._allow_cross and self._buy_quote.price() >= self._sell_quote.price():
            raise Exception("TwoSidedQuote set to not allow crossed quotes. And bid (%s) >= ask (%s)" %
                            (str(self._buy_quote.price()), str(self._sell_quote.price())))

        if self._sell_quote.market() != self._buy_quote.market():
            raise Exception("In a two sided quote, both quotes must be the same market. Buy %s. Sell %s." %
                            (str(self._buy_quote.market()), str(self._sell_quote.market())))

    def buy_quote(self):
        """
        Gets the buy quote

        :return: Quote. Can be None
        """
        return self._buy_quote

    def sell_quote(self):
        """
        Gets the sell quote

        :return: Quote. Can be None
        """
        return self._sell_quote

    def set_buy_quote(self, buy_quote):
        """
        set the buy quote

        :param buy_quote: Quote
        """
        assert buy_quote is None or isinstance(buy_quote, Quote)
        assert buy_quote is None or buy_quote.side().is_bid(), "If buy quote is not None it's Side must be a bid."
        self._buy_quote = buy_quote
        self._validate_quotes()

    def set_sell_quote(self, sell_quote):
        """
        set the sell quote

        :param sell_quote: Quote
        """
        assert sell_quote is None or isinstance(sell_quote, Quote)
        assert sell_quote is None or sell_quote.side().is_ask(), "If sell quote is not None it's Side must be a ask."
        self._sell_quote = sell_quote
        self._validate_quotes()

    def spread(self):
        """
        get the spread between the buy and sell quotes.

        :return: Decimal. Can be None.
        """
        if self.buy_quote() is None or self.sell_quote() is None:
            return None
        return (self.sell_quote().price().price() - self.buy_quote().price().price()) / self.buy_quote().market().product().mpi()

    def __eq__(self, other):
        return self.buy_quote() == other.buy_quote() and self.sell_quote() == other.sell_quote()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "Two Sided Quote: Buy %d (%d visible, %d hidden) at %s. Sell %d (%d visible, %d hidden) at %s." % \
               (self.buy_quote().total_qty(),
                self.buy_quote().visible_qty(),
                self.buy_quote().hidden_qty(),
                str(self.buy_quote().price()),
                self.sell_quote().total_qty(),
                self.sell_quote().visible_qty(),
                self.sell_quote().hidden_qty(),
                str(self.sell_quote().price())
                )



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.Events.OrderEvents import CancelReplaceCommand
from buttonwood.MarketObjects.OrderBooks.OrderLevelOrderBookQueries import modified_qty_at_price
from buttonwood.utils.dicts import NDeepDict


class Priority:
    def __init__(self, ticks_from_tob, ticks_from_opposite_tob, size_ahead_at_price):
        self._ticks_from_tob = ticks_from_tob
        self._ticks_from_opposite_tob = ticks_from_opposite_tob
        self._size_ahead_at_price = size_ahead_at_price

    def ticks_from_tob(self):
        return self._ticks_from_tob

    def ticks_from_opposite_tob(self):
        return self._ticks_from_opposite_tob

    def size_ahead_at_price(self):
        return self._size_ahead_at_price

    def better_priority_than(self, other_priority):
        """
        Returns true if this priority is better than the one passed in.

        Priority is decided by next to get filled so less ticks from TOB or same
         ticks from TOB but less size ahead is better priority.

        :param other_priority: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        :return: Bool
        """
        assert isinstance(other_priority,
                          Priority), "other_priority must be instance of MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority"
        return (self.ticks_from_tob() < other_priority.ticks_from_tob()) or \
               (
               self.ticks_from_tob() == other_priority.ticks_from_tob() and self.size_ahead_at_price() < other_priority.size_ahead_at_price())

    def worse_priority_than(self, other_priority):
        """
        Returns true if this priority is worse than the one passed in.
        
        Priority is decided by next to get filled so more ticks from TOB or same
        ticks from TOB but more size ahead is worse priority.
        
        :param other_priority: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        :return: Bool
        """
        assert isinstance(other_priority,
                          Priority), "other_priority must be instance of MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority"
        return (self.ticks_from_tob() > other_priority.ticks_from_tob()) or \
               (
               self.ticks_from_tob() == other_priority.ticks_from_tob() and self.size_ahead_at_price() > other_priority.size_ahead_at_price())

    def further_from_opposite_tob(self, other_priority):
        """
        Returns true if this priority is further from opposite side tob than
         the one passed in.
    
        :param other_priority: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        :return: Bool
        """
        assert isinstance(other_priority,
                          Priority), "other_priority must be instance of MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority"
        return (self.ticks_from_opposite_tob() > other_priority.ticks_from_opposite_tob()) or \
               (
               self.ticks_from_opposite_tob() == other_priority.ticks_from_opposite_tob() and self.size_ahead_at_price() > other_priority.size_ahead_at_price())

    def closer_to_opposite_tob(self, other_priority):
        """
        Returns true if this priority is further from opposite side tob than
         the one passed in.

        :param other_priority: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        :return: Bool
        """
        assert isinstance(other_priority,
                          Priority), "other_priority must be instance of MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority"
        return (self.ticks_from_opposite_tob() < other_priority.ticks_from_opposite_tob()) or \
               (
               self.ticks_from_opposite_tob() == other_priority.ticks_from_opposite_tob() and self.size_ahead_at_price() < other_priority.size_ahead_at_price())


class EventPriorityListener(OrderLevelBookListener, OrderEventListener):
    """

    priority from event on commands can be considered intended priority

    priority from event on execution reports can be considered achieved priority

    priority at command is what the priority was before the event .

    Designed to work with the order books of multiple markets.
    """

    # TODO UNIT TEST!
    def __init__(self, logger, handle_market_orders=False):
        OrderLevelBookListener.__init__(self, logger)
        OrderEventListener.__init__(self, logger)
        self._market_to_order_book = {}
        self._market_to_event_to_priority = NDeepDict(depth=2)
        self._market_to_event_to_priority_before = NDeepDict(depth=2, default_value=None)
        self._handle_market_orders = handle_market_orders

    def _calculate_priority_not_in_book(self, price, side, market, ignore_order_ids=set()):
        """
        Calculate the priority when not reflected in the book yet.

        :param price: MarketObjects.Price.Price
        :param side: MarketObjects.Side.Side
        :param market: MarketObjects.Market.Market
        :return: PriorityListeners.Priority
        """
        # if the order book is None then the orderbook hasn't been established
        #  and we can assume that ack is best priority for its side and there is
        #  other side.
        order_book = self._market_to_order_book.get(market)
        if order_book is None:
            return Priority(0, None, 0)
        else:  # if order book is not None then we need to figure out what priority would be
            # calculate distance from opposite side since needed in both
            opposite_best_price = order_book.best_price(side.other_side())
            ticks_from_opposite_tob = None
            if opposite_best_price is not None:  # if it is None then ticks from opposite is None
                ticks_from_opposite_tob = price.ticks_behind(opposite_best_price, side, market.product())

            #best_price = order_book.best_price(side)
            best_price = self._best_price(order_book, side, ignore_order_ids)
            # if the best price is None then it is best priority and only need distance from opposite side of book
            if best_price is None:
                return Priority(0, ticks_from_opposite_tob, 0)
            else:  # if best price is not None then we need to calculate
                ticks_from_tob = price.ticks_behind(best_price, side, market.product())
                # assuming visible qty gets priority over hidden so only want visible qty for qty ahead
                if len(ignore_order_ids) == 0:
                    qty_ahead = order_book.visible_qty_at_price(side, price)
                else:
                    qty_ahead = modified_qty_at_price(order_book, side, price, ignore_order_ids=ignore_order_ids,
                                                  ignore_hidden=True)
                return Priority(ticks_from_tob, ticks_from_opposite_tob, qty_ahead)

    def _best_price(self, order_book, side, order_chain_ids):
        prices = order_book.prices(side)
        best_price = None
        for price in prices:
            for order_chain in order_book.iter_order_chains_at_price(side, price):
                if order_chain.chain_id() not in order_chain_ids:
                    best_price = price
                    break
            if best_price is not None:
                break
        return best_price

    def _calculate_priority_in_book(self, order_chain):
        """
        Calculate the priority when is in the book already.

        :param order_chain: MarketObjects.Events.EventChains.OrderEventChain
        :return: PriorityListeners.Priority
        """
        market = order_chain.market()
        side = order_chain.side()
        price = order_chain.current_price() if order_chain.is_open() else order_chain.price_at_close()
        order_book = self._market_to_order_book.get(market)
        if order_book is None:
            return Priority(0, None, 0)
        else:  # if order book is not None then we need to figure out what priority would be
            # calculate distance from opposite side since needed in both
            opposite_best_price = order_book.best_price(side.other_side())
            ticks_from_opposite_tob = None
            if opposite_best_price is not None:  # if it is None then ticks from opposite is None
                ticks_from_opposite_tob = price.ticks_behind(opposite_best_price, side, market.product())

            # in getting best price, ignore the same order chain so it doesn't factor in with cancel-replace back off of top of book.
            best_price = order_book.best_price(side)
            # if the best price is None then it is best priority and only need distance from opposite side of book
            if best_price is None:
                return Priority(0, ticks_from_opposite_tob, 0)
            else:  # if best price is not None then we need to calculate
                ticks_from_tob = price.ticks_behind(best_price, side, market.product())
                # assuming visible qty gets priority over hidden so only want visible qty for qty ahead
                qty_ahead = 0
                for chain in order_book.iter_order_chains_at_price(side, price):
                    if chain.chain_id() == order_chain.chain_id():
                        break
                    qty_ahead += chain.visible_qty()
                return Priority(ticks_from_tob, ticks_from_opposite_tob, qty_ahead)

    def handle_new_order_command(self, new_order_command, resulting_order_chain):
        if new_order_command.is_market_order() and not self._handle_market_orders:
            self._logger.debug("%s set to ignore market orders. New Order %s is a market order. Ignoring" %
                               (self.__class__.__name__, str(new_order_command.event_id())))
            return
        event_id = new_order_command.event_id()
        price = new_order_command.price()
        side = resulting_order_chain.side()
        market = resulting_order_chain.market()
        # new orders are not in the book so calculate what priority *would be*
        priority = self._calculate_priority_not_in_book(price, side, market)
        self._market_to_event_to_priority.set((market, event_id), value=priority)
        # No need to set priority before event on new orders

    def handle_cancel_replace_command(self, cancel_replace_command, resulting_order_chain):
        event_id = cancel_replace_command.event_id()
        price = cancel_replace_command.price()
        side = resulting_order_chain.side()
        market = resulting_order_chain.market()
        # need the most recently requested exposure, if none, get the ack'd exposure
        exposure = resulting_order_chain.most_recent_requested_exposure()
        if exposure is None:
            exposure = resulting_order_chain.current_exposure()

        before_event_priority = self._calculate_priority_in_book(resulting_order_chain)
        self._market_to_event_to_priority_before.set((market, event_id), value=before_event_priority)

        # cancel_replace_same_priority is True if cancel replace down and same price
        cancel_replace_same_priority = (price == exposure.price() and exposure.qty() > cancel_replace_command.qty())

        # cancel replaces have priority at time of event, because order already in book
        if cancel_replace_same_priority:
            priority = before_event_priority
        else:  # otherwise, have to calculate what priority *would be*
            # ignore itself so that we do include it as in front of itself on cancel replace up in size
            priority = self._calculate_priority_not_in_book(price, side, market,
                                                            ignore_order_ids={cancel_replace_command.chain_id()})
        self._market_to_event_to_priority.set((market, event_id), value=priority)

    def handle_cancel_command(self, cancel_command, resulting_order_chain):
        """
        at time fo a cancel, the order should be in the book so just go ahead and calculate priority in book.
        """
        event_id = cancel_command.event_id()
        market = resulting_order_chain.market()
        priority = self._calculate_priority_in_book(resulting_order_chain)
        self._market_to_event_to_priority.set((market, event_id), value=priority)
        # priority before the event is the same calculated aftet event for cancel command
        self._market_to_event_to_priority_before.set((market, event_id), value=priority)

    def handle_acknowledgement_report(self, acknowledgement_report, resulting_order_chain):
        event_id = acknowledgement_report.event_id()
        market = resulting_order_chain.market()
        # acks are in the book already and no need to do anything fancy with ignoring orders
        priority = self._calculate_priority_in_book(resulting_order_chain)
        self._market_to_event_to_priority.set((market, event_id), value=priority)

        # if acking a new order no priority before, so use default of None.
        #  Calculate for cancel replace with current priority since hasn't been applied to book yet
        if isinstance(acknowledgement_report.causing_command(), CancelReplaceCommand):
            self._market_to_event_to_priority_before.set((market, event_id), value=priority)


    def _priority_at_fill(self, fill_event, resulting_order_chain):
        """
        A fill we can assume priority at fill was best.

        Still need to calculate the ticks from opposite side though.
        """
        market = fill_event.market()
        event_id = fill_event.event_id()
        order_book = self._market_to_order_book.get(market)
        # do closes
        opposite_best_price = order_book.best_price(resulting_order_chain.side().other_side())
        ticks_from_opposite_tob = None
        if opposite_best_price is not None:  # if it is None then ticks from opposite is None
            ticks_from_opposite_tob = abs((opposite_best_price - fill_event.fill_price()) / market.product().mpi())
        priority = Priority(0, ticks_from_opposite_tob, 0)
        self._market_to_event_to_priority.set((market, event_id), value=priority)

        #for a fill, priority before fill is always 0
        self._market_to_event_to_priority_before.set((market, event_id), value=priority)

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        self._priority_at_fill(partial_fill_report, resulting_order_chain)

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        self._priority_at_fill(full_fill_report, resulting_order_chain)

    def handle_cancel_report(self, cancel_report, resulting_order_chain):
        """
        When an order is cancelled we want to calculate what its priority was at
         time of cancel. Since the cancel hasn't impacted the order book yet,
         can find the order chain in the order book and calculate the priority.
        """
        # we really only care if a limit and FAR order because otherwise, it wouldn't have any priority since never in book.
        #  and if it has no ack then it was cancelled for self trade purposes (or some such thing) so no impact on book
        if not (resulting_order_chain.is_far() and resulting_order_chain.is_limit_order() and resulting_order_chain.has_acknowledgement()):
            return
        market = resulting_order_chain.market()
        event_id = cancel_report.event_id()
        priority = self._calculate_priority_in_book(resulting_order_chain)
        self._market_to_event_to_priority.set((market, event_id), value=priority)
        self._market_to_event_to_priority_before.set((market, event_id), value=priority)

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        All that needs to be done here is save off the most order_book to be used
         when an event listener call back needs it.
        """
        self._market_to_order_book[order_book.market()] = order_book

    def event_priority(self, market, event_id):
        """
        Get's the priority of the event. If the event was not tracked (/doesn't exist) then will return None.
        
        :param market: MarketObjects.Market.Market
        :param event_id: unique identifier of event
        :return: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        """
        return self._market_to_event_to_priority.get((market, event_id))

    def priority_before_event(self, market, event_id):
        """
        Gets the priority of the event's order chain right before the event occurred. If the event was not 
         tracked (/doesn't exist) then will return None.
        
        :param market: MarketObjects.Market.Market
        :param event_id: unique identifier of event
        :return: MarketMetrics.OrderLevelBookListeners.PriorityListeners.Priority
        """
        return self._market_to_event_to_priority_before.get((market, event_id))

    def clean_up(self, order_chain):
        """
        If clean_up is called with an order_chain than this will go through the
         order chain's events and remove them from tracking

        WARNING: once this is called, the event_priority no longer return the values that are
         meaningful; rather, you'll get None.

        :param order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        market = order_chain.market()
        events = order_chain.events()
        for event in events:
            self._market_to_event_to_priority.delete((market, event.event_id()))
            if event.event_id() in self._market_to_event_to_priority_before.get([market]):
                self._market_to_event_to_priority_before.delete((market, event.event_id()))



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

class OrderLevelBookListener(object):

    def __init__(self, logger):
        self._logger = logger

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        This is a stub to be filled in by each implementing inheriting class.

        The causing_order_chain should have already had the the order event that
        caused the book to update applied to it.

        :param order_book: MarketStructures.OrderBooks.OrderLevelOrderBook.OrderLevelOrderBook
        :param causing_order_chain: MarketStructures.Events.EventChains.OrderEventChain
        :param tob_updated: boolean. Whether or not the top of book updated. This is for speed/convenience for the listener.
        """
        raise NotImplemented("notify_book_update to be implemented by inheriting class.")

    def clean_up_order_chain(self, order_chain):
        """
        Function let's the order book listener clean up data it might be storing for the order chain.
        
        Should only be called if the data being kept by the listener for the order chain (or events in the order chain
         is no longer needed)
         
        :param order_chain: Buttonwood.MarketObjects.Events.EventChains.OrderEventChain
        :return: 
        """
        raise NotImplemented("clean_up_order_chain to be implemented by inheriting class.")



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.MatchSeries import MatchSeries
from buttonwood.utils.dicts import NDeepDict
from collections import defaultdict


class AggressiveAct(MatchSeries):

    def __init__(self, match_id):
        MatchSeries.__init__(self, match_id)
        self._total_aggressive_qty = 0
        self._impact = None

    def aggressing_user_id(self):
        return self._aggressor.user_id()

    def aggressive_side(self):
        return self._aggressor.side()

    def requested_qty(self):
        return self._aggressor.qty()

    def calculate(self, order_book):
        if self.balanced_match_qty():
            impact = 0.0
            passive_side = self._aggressor.side().other_side()
            passive_tob = order_book.best_level(passive_side)
            # if the opposite tob is worse than the last fill price, then there is no impact on top of book. This can
            #  happen at venues that have stupidly bad self trade prevention.
            for fill_price, fill_qty in self._agg_price_to_qty.iteritems():
                if passive_tob is not None and fill_price == passive_tob.price():
                    total_qty = passive_tob.total_qty() + fill_qty
                    impact += float(fill_qty) / total_qty
                elif passive_tob is None or fill_price.better_than(passive_tob.price(), passive_side):
                    impact += 1
            self._impact = impact

        else:
            raise Exception("Cannot calculated aggressive impact because match qty not balanced.")

    def is_closed(self):
        return self._impact is not None

    def add_fill(self, fill_event):
        MatchSeries.add_fill(self, fill_event)

    def impact(self):
        return self._impact


class AggressiveImpactListener(OrderLevelBookListener, OrderEventListener):
    """
    Tracks the Aggressive Impact of each subchain.

    The aggressive impact of a subchain is how many levels of the order book
     that the subchain took out upon open. The left of the decimal is how many
     entire price levels the subchain took out and the right of the decimal is
     the percentage of the next level.

    Aggressive impact value examples:
      * 2 and a half levels --> 2.5
      * 80% of top of book only --> 0.8
      * 4 levels exactly --> 4.0
      * doesn't aggress --> 0.0
    """

    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        OrderEventListener.__init__(self, logger)
        self._market_event_id_aggressive_act = NDeepDict(depth=2, default_value=None)
        self._market_to_orderbook = defaultdict(lambda: None)
        self._market_to_agg_acts_to_close = defaultdict(set)

    def handle_acknowledgement_report(self, acknowledgement_report, resulting_order_chain):
        event_id = acknowledgement_report.causing_command().event_id()

        market = acknowledgement_report.market()
        agg_event = self._market_event_id_aggressive_act.get((market, event_id))
        #if aggressor is acked then all fills on both sides shoud be done and we can calculate
        if agg_event:
            agg_event.calculate(self._market_to_orderbook[market])

    def _handle_fill(self, fill, resulting_order_chain):
        match_id = fill.match_id()
        market = fill.market()
        event_id = fill.causing_command().event_id()

        agg_act = self._market_event_id_aggressive_act.get((market, event_id))
        # if the event_id of the aggressor does not already exist, we create it
        if not agg_act:
            agg_act = AggressiveAct(match_id)
            self._market_event_id_aggressive_act.set((market, event_id), value=agg_act)
        agg_act.add_fill(fill)

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        self._handle_fill(partial_fill_report, resulting_order_chain)

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        self._handle_fill(full_fill_report, resulting_order_chain)
        market = full_fill_report.market()
        # on an aggressive full fill we close out the open order chain
        if full_fill_report.is_aggressor():
            event_id = full_fill_report.causing_command().event_id()
            agg_act = self._market_event_id_aggressive_act.get((market, event_id))
            if agg_act is not None:
                # if full fill and qty balanced we should be able to do the impact calculation right now because book up to date
                if agg_act.balanced_match_qty():
                    agg_act.calculate(self._market_to_orderbook[market])
                else: # we need to get the order level books after all the updates are done.
                    self._market_to_agg_acts_to_close[market].add(self._market_event_id_aggressive_act.get((market, event_id)))
            else:
                raise Exception("Got an aggressive full fill but not tracking aggressive acts for event: %s" % str(event_id))

    def handle_cancel_report(self, cancel_report, resulting_order_chain):
        # on a cancel we close out the open order chain because of FAKs
        market = cancel_report.market()
        event_id = cancel_report.causing_command().event_id()
        agg_event = self._market_event_id_aggressive_act.get((market, event_id))
        # if aggressor is cancelled then all fills on both sides should be done and we can calculate
        if agg_event:
            agg_event.calculate(self._market_to_orderbook[market])

    def clean_up(self, order_chain):
        """
        If clean_up is called with an order_chain than this will go through the
         order chain's subchain IDs and delete each one from the maps that store
         the aggressive acts for a subchain.

        WARNING: once this is called, the get_aggressive_impact and
         get_aggressive_qty will no longer return the values that are
         meaningful; rather, you'll get the 0 values

        :param order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        market = order_chain.market()
        for event in order_chain.events():
            self._market_event_id_aggressive_act.delete([market,event.event_id()])

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        market = order_book.market()
        self._market_to_orderbook[market] = order_book
        remove_set = set()
        agg_acts_to_close = self._market_to_agg_acts_to_close[market]
        for agg_act in agg_acts_to_close:
            if agg_act.balanced_match_qty():
                agg_act.calculate(order_book)
        self._market_to_agg_acts_to_close[market] = agg_acts_to_close.difference(remove_set)

    def get_aggressive_impact(self, market, event_id):
        if self._market_event_id_aggressive_act.get((market, event_id)) is not None:
            impact = self._market_event_id_aggressive_act.get((market, event_id)).impact()
            return impact
        return 0.0

    def get_aggressive_qty(self, market, event_id):
        if self._market_event_id_aggressive_act.get((market, event_id)) is not None:
            return self._market_event_id_aggressive_act.get((market, event_id)).match_qty()
        return 0



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.Events.OrderEvents import FillReport
from buttonwood.MarketObjects.Events.OrderEvents import FullFillReport
from collections import defaultdict


class MatchSeries:
    """
    A match series is a series of fills on both sides of the book that make up one match from one aggressing order.
    
    For example, a 20 lot buy could cross into the offer and match with 4 other resting orders. This class tracks 
     bundles all the resulting fills into one series of matches.
     
    MatchSeries has no requirement that the all the resulting fills are perfectly symmetric because not all venues 
     return symmetric fills. For the above example, one venue might return:
       * Buy 20
       * Sell 7
       * Sell 3
       * Sell 4
       * Sell 6
       
     While another venue returns:
       * Buy 7
       * Buy 3
       * Buy 4
       * Buy 6
       * Sell 7
       * Sell 3
       * Sell 4
       * Sell 6
     
     The only requirement when sanity checks are run on the MatchSeries is that the total size per price are equal
      on both sides of the match.
      
     Also, it uses internal functions for checking that they all equal get_buys(price) get_sells(price) to do the 
      checks. This let's people overwrite those functions in order to have "fuzzy matches" across asset classes or
      products
    
    Match time (timestamp) is time of first fill.
    
    :param match_series_id: unique identifier of the match series
    """

    def __init__(self, match_series_id):
        self._aggressor = None
        self._agg_price_to_qty = defaultdict(int)
        self._pas_price_to_qty = defaultdict(int)
        self._agg_fills = []
        self._pas_fills = []
        self._series_id = match_series_id
        self._match_time = None
        self._agg_fully_filled = False

    def series_id(self):
        return self._series_id

    def timestamp(self):
        return self._match_time

    def add_fill(self, fill_event):
        """
        Adds the passed in fill_event to the MatchSeries.
        
        :param fill_event: Buttonwood.MarketObject.Events.OrderEvents.FillEvent 
        :return: 
        """
        assert (fill_event is not None)
        assert (isinstance(fill_event, FillReport)), "%s is not a FillReport" % type(fill_event)
        assert (fill_event.aggressing_command() is not None)
        assert (fill_event.match_id() == self.series_id()), "The FillEvent's match id needs to match the series' match id"

        # only check the aggressor if it isn't the first one
        if self._aggressor is None:
            self._aggressor = fill_event.aggressing_command()

        if self._match_time is None:
            self._match_time = fill_event.timestamp()

        if fill_event.is_aggressor():
            assert (fill_event.side() == self._aggressor.side()), "An aggressive fill should have same side as aggressor event."
            self._agg_price_to_qty[fill_event.fill_price()] += fill_event.fill_qty()
            self._agg_fills.append(fill_event)
            if isinstance(fill_event, FullFillReport):
                self._agg_fully_filled = True
        else:
            assert (fill_event.side() != self._aggressor.side()), "A passive fill should have different side than aggressor event."
            self._pas_price_to_qty[fill_event.fill_price()] += fill_event.fill_qty()
            self._pas_fills.append(fill_event)

    def aggressor_side(self):
        return None if self._aggressor is None else self._aggressor.side()

    def _price_qty_sanity_check(self):
        # should have same passive and aggressive prices
        if set(self._agg_price_to_qty.keys()) != set(self._pas_price_to_qty.keys()):
            raise Exception("Match Series {}: does not have same prices. Agg: {}. Pas: {}"
                            .format(str(self.series_id()), str(self._agg_price_to_qty.keys()),
                                    str(self._pas_price_to_qty.keys())))

        # should have same qty at each price
        for price, qty in self._agg_price_to_qty.iteritems():
            if qty != self._pas_price_to_qty[price]:
                raise Exception("Match Series %s: does not have same qty at price %s. Agg: %d. Pas: %d" %
                                (str(self.series_id()), str(price), qty, self._pas_price_to_qty[price]))

    def _fills_sanity_check(self):
        # should only have one full fill on aggressive side
        full_fills = []
        if len(self._agg_fills) == 0:
            raise Exception("Match Series %s: has no aggressive fills." % str(self.series_id()))
        if len(self._pas_fills) == 0:
            raise Exception("Match Series %s: has no passive fills." % str(self.series_id()))
        for fill in self._agg_fills:
            if isinstance(fill, FullFillReport):
                full_fills.append(fill)
        if len(full_fills) > 1:
            raise Exception("Match Series {}: has more than one aggressive full fill. {}"
                            .format(str(self.series_id()), str(full_fills)))

    def qty_at_price(self, price, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        return self._agg_price_to_qty[price]

    def price_to_qty(self, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        return self._agg_price_to_qty

    def match_qty(self, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        return self.aggressive_qty()

    def aggressive_qty(self):
        qty = 0
        for fill in self._agg_fills:
            qty += fill.fill_qty()
        return qty

    def passive_qty(self):
        qty = 0
        for fill in self._pas_fills:
            qty += fill.fill_qty()
        return qty

    def balanced_match_qty(self):
        return self.aggressive_qty() == self.passive_qty()

    def prices(self, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        return self._agg_price_to_qty.keys()

    def average_price(self, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        total_qty = 0
        price_weighted_qty = 0.0
        for price, qty in self._agg_price_to_qty.iteritems():
            total_qty += qty
            price_weighted_qty += float(price * qty)
        return price_weighted_qty / total_qty

    def price_in_series(self, price, sanity_check=True):
        if sanity_check:
            self._price_qty_sanity_check()
        return price in self._agg_price_to_qty

    def aggressive_fills(self):
        return self._agg_fills

    def aggressor_fully_filled(self):
        return self._agg_fully_filled

    def passive_fills(self):
        return self._pas_fills

    def aggressing_command(self):
        return self._aggressor

    def run_sanity_checks(self):
        self._fills_sanity_check()
        self._price_qty_sanity_check()



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class Endpoint(object):

    def __init__(self, name, abbreviation=None):
        """
        Contains all the information needed about an Endpoint.

        An endpoint is the *point* where an order *ends*. Some synonyms might be:
         * Liquidity pool
         * ECN
         * Exchange
         * Venue

        The name should be completely unique value for identifying the endpoint. It will be used as an identifying hash.

        The abbreviation can be whatever the user prefers but will be more helpful if unique among endpoints

        A product + an endpoint == a trade-able asset

        :param name: str. the name of the endpoint.
        :param abbreviation: str. an abbreviation of the endpoint. Optional. Defaults to name.
        """
        assert isinstance(name, str)
        assert abbreviation is None or isinstance(abbreviation, str)
        self._name = name
        self._abbr = abbreviation if abbreviation is None else name
        self.__hash = hash(name)

    def name(self):
        """
        Get the name of the Endpoint

        :return: str.
        """
        return self._name

    def abbreviation(self):
        """
        Get the abbreviation of the Endpoint

        :return:
        """
        return self._abbr

    def __eq__(self, other):
        return isinstance(other, Endpoint) and other.__hash__() == self.__hash__()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%s (%s)" % (self.name(), self.abbreviation())

    def to_json(self):
        return {"name": self.name(), "abbreviation": self.abbreviation()}

    def __hash__(self):
        return self.__hash



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2017 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import defaultdict
from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener


class VolumeTracker(object):

    def __init__(self):
        self._counterparty_to_passive_volume = defaultdict(lambda: 0)
        self._counterparty_to_aggressive_volume = defaultdict(lambda: 0)

    def add_passive_trade(self, volume, counterparty):
        self._counterparty_to_passive_volume[counterparty] += volume

    def add_aggressive_trade(self, volume, counterparty):
        self._counterparty_to_aggressive_volume[counterparty] += volume

    def total_aggressive_volume_with(self, counterparty):
        return self._counterparty_to_aggressive_volume[counterparty]

    def total_passive_volume_with(self, counterparty):
        return self._counterparty_to_passive_volume[counterparty]

    def total_volume_with(self, counterparty):
        return self.total_passive_volume_with(counterparty) + self.total_aggressive_volume_with(counterparty)

    def total_aggressive_volume(self):
        total_volume = 0
        for volume in self._counterparty_to_aggressive_volume.itervalues():
            total_volume += volume
        return total_volume

    def total_passive_volume(self):
        total_volume = 0
        for volume in self._counterparty_to_passive_volume.itervalues():
            total_volume += volume
        return total_volume

    def total_volume(self):
        return self.total_passive_volume() + self.total_aggressive_volume()


class VolumeTrackingListener(OrderEventListener):

    def __init__(self, logger):
        OrderEventListener.__init__(self, logger)
        self._market_to_participant_to_volume = defaultdict(lambda: defaultdict(VolumeTracker()))  # TODO can I use NDeep dict here?

    def _handle_fill(self, fill_report):
        # only care about passive fill reports because that way we get both counterparties:
        #  1) the user getting filled here and
        #  2) the user on the aggressive command
        if not fill_report.is_aggressor():
            passive_user = fill_report.user()
            aggressive_user = fill_report.aggressing_command().user()
            qty = fill_report.fill_qty()
            market = fill_report.market()
            self._market_to_participant_to_volume[market][passive_user].add_passive_trade(qty, aggressive_user)
            self._market_to_participant_to_volume[market][aggressive_user].add_aggressive_trade(qty, passive_user)

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        self._handle_fill(partial_fill_report)

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        self._handle_fill(full_fill_report)
